__title__ = "trino"
__description__ = "Client for the Trino distributed SQL Engine"
__url__ = "https://github.com/trinodb/trino-python-client"
__version__ = "0.332.0"
__author__ = "Trino Team"
__author_email__ = "python-client@trino.io"
__license__ = "Apache 2.0"
